# Generated by Django 5.0.1 on 2024-01-30 19:43

from django.db import migrations, models

import django_fast_iprestrict.validators


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0009_alter_ruleratelimit_rate"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ruleratelimit",
            name="rate",
            field=models.CharField(
                help_text=(
                    'Set to "inherit" to use the provided rate (when set to "inherit" '
                    "ratelimit is ignored in middleware and without a provided rate)"
                ),
                max_length=10,
                validators=[django_fast_iprestrict.validators.validate_rate],
            ),
        ),
    ]
