# Generated by Django 5.0.1 on 2024-01-30 19:05

import django_fast_iprestrict.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fast_iprestrict', '0008_add_managed_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ruleratelimit',
            name='rate',
            field=models.CharField(help_text='Leave empty to use the default rate (only valid for ratelimit matching rules)', max_length=10, validators=[django_fast_iprestrict.validators.validate_rate]),
        ),
    ]
