# Generated by Django 5.0.2 on 2024-02-07 11:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fast_iprestrict', '0010_alter_ruleratelimit_rate'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='ruleratelimitgroup',
            name='rule_ratelimit_groups_unique',
        ),
        migrations.AddConstraint(
            model_name='ruleratelimitgroup',
            constraint=models.UniqueConstraint(fields=('rule', 'name'), name='rule_ratelimit_groups_unique'),
        ),
    ]
