# Generated by Django 5.0.1 on 2024-01-03 11:12

import django_fast_iprestrict.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0005_add_ratelimit_fix_generator_fn"),
    ]

    operations = [
        migrations.AddField(
            model_name="rule",
            name="invert_methods",
            field=models.BooleanField(
                blank=True,
                default=True,
                help_text="exclude instead of include http methods. Set and leave methods empty to match all",
            ),
        ),
        migrations.AddField(
            model_name="rule",
            name="methods",
            field=models.CharField(
                blank=True,
                default="",
                help_text="comma seperated http methods",
                max_length=100,
                validators=[django_fast_iprestrict.validators.validate_methods],
            ),
        ),
        migrations.AddField(
            model_name="ruleratelimit",
            name="action",
            field=models.SmallIntegerField(
                choices=[
                    (2, "normal"),
                    (1, "annotate only"),
                    (3, "clear cache for key"),
                    (4, "reset to before request"),
                ],
                default=2,
            ),
        ),
    ]
